import ast
import textwrap

def extract_function_from_file(file_path: str, function_name: str) -> str:
    with open(file_path, 'r') as f:
        source = f.read()

    parsed_ast = ast.parse(source)
    for node in ast.walk(parsed_ast):
        if isinstance(node, ast.FunctionDef) and node.name == function_name:
            lines = source.splitlines()
            start_line = node.lineno - 1
            end_line = node.end_lineno
            func_lines = lines[start_line:end_line]
            return textwrap.dedent("\n".join(func_lines))

    raise ValueError(f"Function '{function_name}' not found in {file_path}")



import sys
from pathlib import Path
import runpy
import os
from aipymemtimeprofiler.config.config import file_path,dir_path,console_display,agentic_profiler
from aipymemtimeprofiler.profiler.profile_details import Profiler
from aipymemtimeprofiler.flask_app_profiler.load_flask import load_flask_app,wrap_flask_routes
from aipymemtimeprofiler.analyser.performance_analyser import analyse_performance,collect_profiling_data
from aipymemtimeprofiler.flask_app_profiler.profiler import display_functions_and_select
from rich import print
from rich.console import Console
from rich.table import Table



def run_with_profiler(leak_threshold_kb=10):
    console = Console()
    app, app_dir = load_flask_app(file_path)
    if app:
        wrap_flask_routes(app, app_dir)
        app.run(debug=True)
        if agentic_profiler == True:
            profiling_data = collect_profiling_data()
            selected_func = display_functions_and_select(profiling_data)
            if selected_func != None:
                print("\n[Selected] Function:", selected_func[0])
                print("[Selected] File Path:", selected_func[1])
                analyse_result = analyse_performance(selected_func[1],selected_func[0])
                print(analyse_result)
    else:
        filepath = Path(file_path).resolve()
        console_display= True
        if console_display == 'False':
            console_display = False
        else:
            console_display = True
        profiler = Profiler(dir_path,console_display,leak_threshold_kb=leak_threshold_kb)
        
        prev_cwd = os.getcwd()
        sys.setprofile(profiler.profile_func)
        print(dir_path)
        os.chdir(dir_path)
        print(file_path)
        sys.path.insert(0, str(dir_path))
        try:
            runpy.run_path(str(filepath), run_name="__main__")
        finally:
            sys.setprofile(None)
            os.chdir(prev_cwd)
            profiler.write_output()
            print("ENABLE AI : ",agentic_profiler)
            if agentic_profiler == True:
                profiling_data = collect_profiling_data()
                selected_func = display_functions_and_select(profiling_data)
                if selected_func != None:
                    print("\n[Selected] Function:", selected_func[0])
                    print("[Selected] File Path:", selected_func[1])
                    analyse_result = analyse_performance(selected_func[1],selected_func[0])
                    print(analyse_result)
            

if __name__ == "__main__":
    run_with_profiler()


import sys
import time
import tracemalloc
import json
from pathlib import Path
import os
import gc
from pympler import asizeof
import psutil
import io
from line_profiler import LineProfiler
import re
from rich import print
from rich.console import Console
from rich.table import Table

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, ROOT_DIR)
from aipymemtimeprofiler.helpers.helper import check_event, check_external_functions, get_all_args, is_user_code, detect_mem_leak,check_external_lib, get_package
from aipymemtimeprofiler.config.config import file_path, dir_path,INCLUDE_LIBRARIES

class Profiler:
    def __init__(self, root_path,console_display=True, leak_threshold_kb=3000):
        self.root_path = Path(root_path).resolve()
        self.records = {}
        self.call_stack = {}
        self.leak_threshold_kb = leak_threshold_kb
        self.proc = psutil.Process(os.getpid())
        self.console_display = console_display
        self.external_libraries = {}
        tracemalloc.start()

        self.line_profiler = LineProfiler()
        self.line_profiled_functions = set()
        self.line_profiled_output = {}
        self.enable_line_profiling = True

    def profile_func(self, frame, event, arg):
        external_lib = False
        if check_event(event):
            return self.profile_func

        code = frame.f_code
        func_name = code.co_name

        if check_external_functions(func_name,code.co_filename):
            if INCLUDE_LIBRARIES and check_external_lib(code.co_filename):
                external_lib = True
            else:
                return self.profile_func
                



        # if not is_user_code(frame, self.root_path):
        #     return self.profile_func

        key = (func_name, code.co_filename, frame.f_lineno)

        if event == "call":
            gc.collect()

            all_args = get_all_args(frame)
            key = (func_name, code.co_filename, frame.f_lineno,' '.join(all_args.keys()))
            cpu_start = time.process_time()
            wall_start = time.perf_counter()
            start_snapshot = tracemalloc.take_snapshot()
            mem_start = self.proc.memory_info().rss

            self.call_stack[frame] = {
                "start_time": wall_start,
                "start_cpu": cpu_start,
                "start_snapshot": start_snapshot,
                "args": all_args,
                "mem_start": mem_start
            }

            # print(f"[CALL] {func_name} at {code.co_filename}")

        elif event == "return" and frame in self.call_stack:
            all_args = get_all_args(frame)
            key = (func_name, code.co_filename, frame.f_lineno,' '.join(all_args.keys()))
            code_obj = frame.f_code

            if self.enable_line_profiling and code_obj not in self.line_profiled_functions:
                func = frame.f_globals.get(code_obj.co_name)
                if callable(func):
                    self.line_profiler.add_function(func)
                    self.line_profiled_functions.add(code_obj)

            call_info = self.call_stack.pop(frame)
            end_time = time.perf_counter()
            end_cpu = time.process_time()
            mem_end = self.proc.memory_info().rss

            duration = end_time - call_info["start_time"]
            cpu_time = end_cpu - call_info["start_cpu"]
            gc.collect()
            end_snapshot = tracemalloc.take_snapshot()
            memory_diff = end_snapshot.compare_to(call_info["start_snapshot"], "lineno")
            total_mem = sum([stat.size_diff for stat in memory_diff])
            max_mem = max([stat.size_diff for stat in memory_diff], default=0)
            mem_growth = mem_end - call_info["mem_start"]

            returned_size = asizeof.asizeof(arg)
            returned_enough = abs(total_mem - returned_size) / 1000
            leak_candidates = detect_mem_leak(memory_diff, self.leak_threshold_kb, self.root_path)

            record = self.records.get(
                key,
                {
                    "function": func_name,
                    "file": code.co_filename,
                    "line": frame.f_lineno,
                    "max_time": 0,
                    "cpu_time": 0,
                    "max_mem": 0,
                    "mem_growth_rss": 0,
                    "mem_growth": 0,
                    "args": call_info["args"],
                    "possible_memory_leak": None,
                    "note": []
                },
            )

            record["max_time"] = max(record["max_time"], duration)
            record["cpu_time"] = max(record["cpu_time"], cpu_time)
            record["max_mem"] = max(record["max_mem"], max_mem)
            record["max_time_ms"] = round(record["max_time"] * 1000, 3)
            record["cpu_time_ms"] = round(record["cpu_time"] * 1000, 3)
            record["max_mem_kb"] = round(record["max_mem"] / 1024, 3)
            record["mem_growth_rss"] = mem_growth
            record["mem_growth_rss_kb"] = max(0,round(mem_growth / 1024, 3))
            record["returned_size"] = returned_size

            if (self.leak_threshold_kb * 1000) < returned_size:
                record["note"].append("Obj return size is huge. Please check.")

            if external_lib:
                self.external_libraries[key] = record
            else:
                self.records[key] = record

        return self.profile_func

    def write_output(self, output_file="profile_output.json"):

        output = list(self.records.values())
        with open(output_file, "w") as f:
            json.dump(output, f, indent=2)
        
        if self.console_display:
            print(f"\n[bold green]✅ Profile results written to[/bold green] [cyan]{output_file}[/cyan]")
            console = Console()
            console.rule("[bold yellow]📊 Profiling Summary[/bold yellow]")
            table = Table(show_header=True, header_style="bold magenta", title="Function Profile Overview")
            table.add_column("Function", style="bold cyan")
            table.add_column("Time (ms)", justify="right")
            table.add_column("CPU Time (ms)", justify="right")
            table.add_column("Memory Growth (KB)", justify="right")
            table.add_column("Package", justify="right")
            table.add_column("Note")
            
            output = list(self.external_libraries.values())
            output.extend(list(self.records.values()))

            for record in output:
                note_lines = record.get("note", [])
                if note_lines:
                    note_str = "[bold red]🔴 " + "[/bold red][bright_red]" + "\n      ".join(note_lines) + "[/bright_red]"
                else:
                    note_str = "-"
                if check_external_lib(record['file']):
                    package = get_package(record['file'])
                    table.add_row(
                        record["function"],
                        f"{record['max_time_ms']:.2f}",
                        f"{record['cpu_time_ms']:.2f}",
                        f"{record['mem_growth_rss_kb']:.2f}",
                        package,
                        note_str
                    )
                else:
                    table.add_row(
                        record["function"],
                        f"{record['max_time_ms']:.2f}",
                        f"{record['cpu_time_ms']:.2f}",
                        f"{record['mem_growth_rss_kb']:.2f}",
                        "user_code",
                        note_str
                    )

            console.print(table)



import sys
import os
import importlib.util
import time
import psutil
import types
import inspect
from flask import Flask
from memory_profiler import memory_usage
from rich import print
from rich.console import Console
from rich.table import Table
import json
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, ROOT_DIR)
from aipymemtimeprofiler.config.config import agentic_profiler,INCLUDE_LIBRARIES
from aipymemtimeprofiler.analyser.performance_analyser import collect_profiling_data,analyse_performance
from aipymemtimeprofiler.helpers.helper import check_external_functions,get_all_args,check_external_lib,get_package
process = psutil.Process(os.getpid())

def display_functions_and_select(functions_with_files):
    console = Console()
    table = Table(title="Available Functions for Analysis")

    table.add_column("Index", justify="center", style="cyan", no_wrap=True)
    table.add_column("Function Name", style="magenta")
    table.add_column("File Path", style="green")

    for idx, (func_name, file_path) in enumerate(functions_with_files):
        table.add_row(str(idx), func_name, file_path)

    # Add the "Skip Analysis" option
    skip_index = len(functions_with_files)
    table.add_row(str(skip_index), "[bold yellow]Skip Analysis[/bold yellow]", "-")

    console.print(table)

    while True:
        try:
            selected_index = int(console.input("[bold blue]Enter the index of the function you want to analyze (or skip):[/bold blue] "))
            if 0 <= selected_index < len(functions_with_files):
                return functions_with_files[selected_index]
            elif selected_index == skip_index:
                console.print("[bold yellow]Skipping analysis.[/bold yellow]")
                return None
            else:
                console.print("[red]Invalid index. Try again.[/red]")
        except ValueError:
            console.print("[red]Please enter a valid number.[/red]")

class flask_profiler:
    def __init__(self,console_display= True):
        self.records = {}
        self.console_display = console_display

    def trace_calls(self, frame, event, arg):
        if event != "call":
            return

        code = frame.f_code
        filename = code.co_filename
        func_name = code.co_name
        global_vars = frame.f_globals
        used_names = frame.f_code.co_names
        used_globals = {
            name: global_vars[name] for name in used_names if name in global_vars
        }
        
        # Ignore system libs
        if "site-packages" in filename or "lib/python" in filename:
            return
        if check_external_functions(func_name,code.co_filename):
            return
        if check_external_functions(func_name,code.co_filename):
            if INCLUDE_LIBRARIES and check_external_lib(code.co_filename):
                external_lib = True
            else:
                return

        def local_trace(frame, event, arg,self=self):
            all_args = get_all_args(frame)
            key = (func_name, code.co_filename, frame.f_lineno,' '.join(all_args.keys()))
            if event == "return":
                func = frame.f_globals.get(frame.f_code.co_name)

                start_time = frame.f_locals.get("__start_time__")
                start_cpu = frame.f_locals.get("__start_cpu__")
                start_mem = frame.f_locals.get("__start_mem__")

                if start_time is not None:
                    end_time = time.perf_counter()
                    end_cpu = process.cpu_times().user
                    end_mem = process.memory_info().rss

                    duration = end_time - start_time
                    cpu_time = end_cpu - start_cpu
                    mem_growth = max(0.00,end_mem - start_mem)
                    returned_size = sys.getsizeof(arg)
                    record = self.records.get(
                        key,
                        {
                            "function": func_name,
                            "file": code.co_filename,
                            "line": frame.f_lineno,
                            "max_time": 0,
                            "cpu_time": 0,
                            "max_mem": 0,
                            "mem_growth_rss": 0,
                            "args": all_args,
                            "possible_memory_leak": None,
                            "note": []
                        },
                    )
                    record["function"] = func_name
                    record["file"]= filename
                    record["line"] = frame.f_code.co_firstlineno
                    record["max_time"] = duration
                    record["cpu_time"] = cpu_time
                    record["max_time_ms"] = round(duration * 1000, 3)
                    record["cpu_time_ms"]= round(cpu_time * 1000, 3)
                    record["mem_growth_rss"] = mem_growth
                    record["mem_growth_rss_kb"] = round(mem_growth / 1024, 2)
                    record["returned_size"] = returned_size
                    record["possible_memory_leak"]= None
                    record["note"]= ["Obj return size is huge. Please check."] if returned_size > 10 * 1024 * 1024 else []
                    self.records[key] = record
            elif event == "call":
                frame.f_locals["__start_time__"] = time.perf_counter()
                frame.f_locals["__start_cpu__"] = process.cpu_times().user
                frame.f_locals["__start_mem__"] = process.memory_info().rss
            elif event == "exception":
                all_args = get_all_args(frame)
                key = (func_name, code.co_filename, frame.f_lineno,' '.join(all_args.keys()))
                start_time = frame.f_locals.get("__start_time__")
                start_cpu = frame.f_locals.get("__start_cpu__")
                start_mem = frame.f_locals.get("__start_mem__")
                if start_time is not None:
                    end_time = time.perf_counter()
                    end_cpu = process.cpu_times().user
                    end_mem = process.memory_info().rss
                duration = end_time - start_time
                cpu_time = end_cpu - start_cpu
                mem_growth = end_mem - start_mem
                record = self.records.get(
                    key,
                    {
                        "function": func_name,
                        "file": code.co_filename,
                        "line": frame.f_lineno,
                        "max_time": 0,
                        "cpu_time": 0,
                        "max_mem": 0,
                        "mem_growth_rss": 0,
                        "args": all_args,
                        "possible_memory_leak": None,
                        "note": []
                    },
                )
                record["function"] = func_name
                record["file"]= filename
                record["max_time"]= duration
                record["cpu_time"]= cpu_time
                record["max_time_ms"] = round(duration * 1000, 3)
                record["cpu_time_ms"] = round(cpu_time * 1000, 3)
                record["note"] = ["Exception occoured in this function"]

                self.records[key] = record
            return local_trace

        frame.f_locals["__start_time__"] = time.perf_counter()
        frame.f_locals["__start_cpu__"] = process.cpu_times().user
        frame.f_locals["__start_mem__"] = process.memory_info().rss

        return local_trace
    
    
    def write_output(self, output_file="profile_output.json"):
        def safe_serialize(obj):
            try:
                json.dumps(obj)
                return obj
            except TypeError:
                return str(obj)

        cleaned_records = [
            {k: safe_serialize(v) for k, v in record.items()}
            for record in self.records.values()
        ]

        with open(output_file, "w") as f:
            json.dump(cleaned_records, f, indent=2)

        output = cleaned_records

        print(f" Wrote {len(cleaned_records)} records to {output_file}")
        if self.console_display:
            print(f"\n[bold green]✅ Profile results written to[/bold green] [cyan]{output_file}[/cyan]")
            console = Console()
            console.rule("[bold yellow]📊 Profiling Summary[/bold yellow]")
            table = Table(show_header=True, header_style="bold magenta", title="Function Profile Overview")
            table.add_column("Function", style="bold cyan")
            table.add_column("Time (ms)", justify="right")
            table.add_column("CPU Time (ms)", justify="right")
            table.add_column("Growth (KB)", justify="right")
            table.add_column("Package", justify="right")
            table.add_column("Note")            

            for record in output:
                print(record)
                note_lines = record.get("note", [])
                if note_lines:
                    print(note_lines)
                    note_str = "[bold red]🔴 " + "[/bold red][bright_red]" + "\n      ".join(note_lines) + "[/bright_red]"
                else:
                    note_str = "-"
                if check_external_lib(record['file']):
                    package = get_package(record['file'])
                    table.add_row(
                        record["function"],
                        f"{record['max_time_ms']:.2f}",
                        f"{record['cpu_time_ms']:.2f}",
                        f"{record['mem_growth_rss_kb']:.2f}",
                        package,
                        note_str
                    )
                else:
                    table.add_row(
                        record["function"],
                        f"{record['max_time_ms']:.2f}",
                        f"{record['cpu_time_ms']:.2f}",
                        f"{record['mem_growth_rss_kb']:.2f}",
                        "user_code",
                        note_str
                    )


            console.print(table)


import sys
import os
import importlib.util
import time
import psutil
import types
import inspect
from flask import Flask
from memory_profiler import memory_usage
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, ROOT_DIR)
from aipymemtimeprofiler.flask_app_profiler.profiler import flask_profiler,display_functions_and_select
from aipymemtimeprofiler.config.config import agentic_profiler
from aipymemtimeprofiler.analyser.performance_analyser import collect_profiling_data,analyse_performance

def load_flask_app(app_path):
    try:
        app_path = os.path.abspath(app_path)
        app_dir = os.path.dirname(app_path)
        module_name = os.path.splitext(os.path.basename(app_path))[0]

        sys.path.insert(0, app_dir)  # allow relative imports
        spec = importlib.util.spec_from_file_location(module_name, app_path)
        module = importlib.util.module_from_spec(spec)  # type: ignore
        sys.modules[module_name] = module
        spec.loader.exec_module(module)  # type: ignore

        for attr in dir(module):
            obj = getattr(module, attr)
            if isinstance(obj, Flask):
                return obj, app_dir
        return None, None
    except Exception as e:
        print(f"Error loading Flask app: {e}")
        return None, None
    # raise Exception("No Flask app found in the given file.")


def wrap_flask_routes(app,dir_path):
    for rule in app.url_map.iter_rules():
        endpoint = app.view_functions[rule.endpoint]
        app.view_functions[rule.endpoint] = wrap_function(endpoint,dir_path)


def wrap_function(func, dir_path):
    def wrapper(*args, **kwargs):
        profiler = flask_profiler()

        # Temporarily change working directory
        prev_dir = os.getcwd()
        os.chdir(dir_path)

        try:
            sys.settrace(profiler.trace_calls)
            mem_used, result = memory_usage((func, args, kwargs), retval=True, max_usage=True, interval=0.01)  # type: ignore
            sys.settrace(None)
            os.chdir(prev_dir)
            profiler.write_output()
        finally:
            os.chdir(prev_dir)

        return result

    wrapper.__name__ = func.__name__
    return wrapper


import os
file_path = os.getenv('PROFILER_FILE_PATH','None')
dir_path = os.getenv('PROFILER_DIR_PATH','None')
console_display = os.getenv('CONSOLE_DISPLAY',True)
agentic_profiler = os.getenv('AGENTIC_PROFILER',False)
OLLAMA_MODEL=os.getenv('AGENT_NAME','deepseek-r1:1.5b')
INCLUDE_LIBRARIES = os.getenv('INCLUDE_LIBRARIES',False)


import ollama
import json
import os
import re
import sys
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, ROOT_DIR)
from aipymemtimeprofiler.analyser.function_extract import extract_function_from_file
from aipymemtimeprofiler.config.config import OLLAMA_MODEL


def connect_to_ollama():
    return ollama

def remove_think_blocks(text):
    text = text.split("content=")[1]
    return re.sub(r"<think>.*?</think>", "", text, flags=re.DOTALL)

# Step 2 (Optional): Make **bold** sections uppercase or add console colors
def emphasize_bold(text):
    return re.sub(r"\*\*(.*?)\*\*", lambda m: m.group(1).upper(), text)

def send_analysis_request(code):
    client = connect_to_ollama()
    prompt = f"""
        You are a Python programming expert.
        You have a keen eye for unused variables, memory and time optimization.
        Ignore the imports. Assume we have proper imports to the function
        Given the following Python code in section __code__ , your task is to analyse the code deeply check if there are absolutely necessary variables be careful while doing this think twice:
        1. **Identify any errors** (e.g., syntax errors, logical mistakes, potential runtime errors).
        2. **Find inefficiencies** in the code (e.g., unnecessary computations, inefficient algorithms, redundant operations).
        3. **Suggest improvements** to fix the errors and optimize the code. Your suggestions should focus on:
            - Correcting any errors.
            - Improving efficiency (both in terms of time and memory).
            - Refactoring the code for better readability and maintainability.
            - Following Python best practices (e.g., variable naming, code style according to PEP 8).
            - Any other performance optimizations or recommendations.
            - Check if there are any unused variables in the code. If there are it's a possible memory leak
            - Check the unused variables twice. Think about this.
            - If there are any global variables used. Warn that global variables are used in this function.

        Do not provide any extra explanation or commentary beyond the identification and suggestions.
        Print only short answers. Give only Analysis and suggested code
        If you don't have any important problems. Then say Everything looks good and don't suggest anything.
        ****problem and analysis****
        Write the problems found and analysis here

        ****suggested_code****
        Suggested code here

        Here is the code to analyze:
        __code__
        {code}
        """
    response = client.chat(model=OLLAMA_MODEL, messages=[{"role": "user", "content": prompt}])
    response = remove_think_blocks(str(response))
    formatted_content = response.encode().decode('unicode_escape')
    readable_output = emphasize_bold(formatted_content)

    # print("Full Response:", response,type(response))

    try:
        return readable_output
    except json.JSONDecodeError:
        print("Failed to return content:", response)
        return None



def collect_profiling_data():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    file_path = os.path.normpath(os.path.join(script_dir, "..", "..", "profile_output.json"))
    if not os.path.exists(file_path):
        print(f"Error: File not found at path '{file_path}'")
        return []

    try:
        with open(file_path, "r") as f:
            profiling_data = json.load(f)

            if not isinstance(profiling_data, list):
                print("Error: JSON root should be a list of profiling entries.")
                return []

            # Extract only (function, file) as a list of tuples
            return [(entry["function"], entry["file"]) for entry in profiling_data if "function" in entry and "file" in entry]

    except json.JSONDecodeError:
        print("Error: Failed to decode JSON.")
        return []
    except Exception as e:
        print(f"Unexpected error: {e}")
        return []



def analyze_performance(profiling_function):
    analysis_results = send_analysis_request(profiling_function)
    
    if analysis_results:
        return analysis_results
    else:
        print("No analysis results returned.")
        return None

def analyse_performance(function_file,function_name):
    code = extract_function_from_file(function_file,function_name)
    print("Analysing code \n", code)
    analysis_results = analyze_performance(code)
    print("Analysis completed")
    return analysis_results


from setuptools import setup, find_packages

setup(
    name="aipymemtimeprofiler",
    version="0.3.0",
    author="Nikhil Lingadhal",
    description="AI-enabled time and memory profiler for Python applications",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/nikhillingadhal1999/AIPyMemTimeProfiler",
    license="MIT",
    packages=find_packages(),  # auto-detects submodules in aipymemtimeprofiler/
    install_requires=[
        "line_profiler==4.2.0",
        "psutil==7.0.0",
        "Pympler==1.1",
        "rich",
        "Flask",
        "memory_profiler",
        "setuptools",
        "wheel",
        "twine",
        "ollama",
        "packaging>=23.2,<25"
    ],
    entry_points={
        "console_scripts": [
            "profile_code = aipymemtimeprofiler.executable.execute:run_with_profiler"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import inspect
from pathlib import Path


def check_event(event):
    return event not in ("call", "return")


def check_external_functions(func_name,file_name):
    return (
        func_name.startswith("<") and func_name.endswith(">")
    ) or "<frozen importlib._bootstrap>" in file_name or "site-packages" in file_name or "lib/python" in file_name or (file_name.startswith("<") and file_name.endswith(">"))

def check_external_lib(file_name):
    return "site-packages" in file_name

def get_package(file_name):
    return file_name.split("site-packages")[1].split("/")[1]

def get_all_args(frame):
    args_info = inspect.getargvalues(frame)
    args_repr = {}
    for k in args_info.args:
        try:
            args_repr[k] = repr(args_info.locals[k])
        except Exception:
            args_repr[k] = "<unrepresentable>"
    return args_repr


def is_user_code(frame_or_path, root_path):
    file_path = Path(
        frame_or_path.f_code.co_filename
        if hasattr(frame_or_path, "f_code")
        else frame_or_path
    ).resolve()
    return str(file_path).startswith(str(root_path))


def detect_mem_leak(memory_diff, leak_threshold_kb, root_path):
    leak_candidates = []
    for stat in memory_diff:
        size_kb = stat.size_diff / 1024
        if (
            stat.size_diff > 0
            and is_user_code(stat.traceback[0].filename, root_path)
            and size_kb >= leak_threshold_kb
        ):
            leak_candidates.append(
                {
                    "file": str(Path(stat.traceback[0].filename).resolve()),
                    "line": stat.traceback[0].lineno,
                    "size_kb": round(size_kb, 3),
                }
            )
    return leak_candidates